# Copyright 2023 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Values for Client V2 in Bluecat libraries."""


class MediaType:
    """Values for the media type in header"""

    CSV = "text/csv"
    FORM_DATA = "multipart/form-data"
    HAL_JSON = "application/hal+json"
    HTML = "text/html"
    JSON = "application/json"
    MERGE_PATCH_JSON = "application/merge-patch+json"
    OCTET_STREAM = "application/octet-stream"
